"""
Translate JADN to/from Markdown tables
"""
import json
import re

from datetime import datetime
from typing import NoReturn, Tuple, Union
from jadn.definitions import TypeName, BaseType, TypeOptions, TypeDesc, Fields, ItemID, FieldID, INFO_ORDER
from jadn.utils import cleanup_tagid, get_optx, fielddef2jadn, jadn2fielddef, jadn2typestr, raise_error, typestr2jadn

# MARKDOWN -> JADN Type regexes
p_tname = r'\s*([-$\w]+)'               # Type Name
p_assign = r'\s*='                      # Type assignment operator
p_tstr = r'\s*(.*?)\s*\{?'              # Type definition
p_tdesc = r'(?:\s*\/\/\s*(.*?)\s*)?'    # Optional Type description

# MARKDOWN -> JADN Field regexes
p_id = r'\s*(\d+)'  # Field ID
p_fname = r'\s+([-:$\w]+\/?)?'  # Field Name with dir/ option (colon is deprecated, allow for now)
p_fstr = r'\s*(.*?)'  # Field definition or Enum value
p_range = r'\s*(?:\[([.*\w]+)\]|(optional))?'  # Multiplicity
p_desc = r'\s*(?:\/\/\s*(.*?)\s*)?'  # Field description, including field name if .id option


def markdown_style() -> dict:
    return {
        'pad': True,    # Use one space horizontal padding
        'links': True   # Retain Markdown links: [text](link)
    }


def markdown_dumps(schema: dict, style: dict = None) -> str:
    """
    Convert JADN schema to Markdown Tables

    :param dict schema: JADN schema
    :param dict style: Override default options if specified
    :return: Markdown tables
    :rtype: str
    """
    w = markdown_style()
    if style:
        w.update(style)   # Override any specified style options

    text = ''
    info = schema['info'] if 'info' in schema else {}
    mlist = [k for k in INFO_ORDER if k in info]
    for k in mlist + list(set(info) - set(mlist)):      # Display info elements in fixed order
        text += f'{k:>14}: {json.dumps(info[k])}\n'     # TODO: wrap to width, continuation-line parser

    for td in schema['types']:
        if len(td) > Fields and td[Fields]:
            tdef = f'{td[TypeName]} ({jadn2typestr(td[BaseType], td[TypeOptions])})'
            tdesc = f'\n{td[TypeDesc]}\n' if td[TypeDesc] else ''
            text += f'{tdesc}\n**Type: ' + tdef.replace("*", "\*") + '**\n'
            idt = td[BaseType] == 'Array' or get_optx(td[TypeOptions], 'id') is not None
            table_type = (0 if td[BaseType] == 'Enumerated' else 2) + (0 if idt else 1)
            table = [
                [['ID', 'Description']],
                [['ID', 'Item', 'Description']],
                [['ID', 'Type', '\#', 'Description']],
                [['ID', 'Name', 'Type', '\#', 'Description']]
            ][table_type]
            for fd in td[Fields]:
                fname, fdef, fmult, fdesc = jadn2fielddef(fd, td)
                fdef = fdef.replace('*', '\*')
                fmult = fmult.replace('*', '\*')
                dsc = fdesc.split('::', maxsplit=2)
                fdesc = f'**{dsc[0]}** - {dsc[1].strip()}' if len(dsc) == 2 else fdesc
                if table_type == 0:
                    table.append([str(fd[ItemID]), fdesc])
                elif table_type == 1:
                    table.append([str(fd[ItemID]), f'**{fname}**', fdesc])
                elif table_type == 2:
                    table.append([str(fd[FieldID]), fdef, fmult, fdesc])
                elif table_type == 3:
                    table.append([str(fd[FieldID]), f'**{fname}**', fdef, fmult, fdesc])
        else:
            table = [['Type Name', 'Type Definition', 'Description'],
                     [f'**{td[TypeName]}**', jadn2typestr(td[BaseType], td[TypeOptions]), td[TypeDesc]]]
        text += f'\n{format_table(table)}\n\n**********\n'
    return text


def format_table(rows: list) -> str:
    cwidth = [len(data.strip()) for data in rows[0]]
    for row in rows[1:]:
        for c in range(len(row)):
            cwidth[c] = max(cwidth[c], len(row[c].strip()))
    hbar = f'|{"|".join(["-" * (c + 2) for c in cwidth])}|'
    cf = f'| {" | ".join(["{:" + str(c) + "}" for c in cwidth])} |'
    return '\n'.join([cf.format(*rows[0])] + [hbar] + [cf.format(*r) for r in rows[1:]])


def markdown_dump(schema: dict, fname: Union[bytes, str, int], source='', style=None) -> NoReturn:
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write(f'<!--- Generated from {source}, {datetime.ctime(datetime.now())} --->\n\n')
        f.write(markdown_dumps(schema, style))


# Convert MARKDOWN to JADN
def line2jadn(line: str, tdef: list) -> Tuple[str, list]:
    if line.split('//', maxsplit=1)[0].strip():
        p_info = r'^\s*([-\w]+):\s*(.+?)\s*$'
        if m := re.match(p_info, line):
            return 'M', [m.group(1), m.group(2)]

        p_type = fr'^{p_tname}{p_assign}{p_tstr}{p_tdesc}$'
        if m := re.match(p_type, line):
            btype, topts, fo = typestr2jadn(m.group(2))
            assert fo == []                     # field options MUST not be included in typedefs
            newtype = [m.group(1), btype, topts, m.group(3) if m.group(3) else '', []]
            return 'T', newtype

        if tdef:        # looking for fields
            pn = '()' if (get_optx(tdef[TypeOptions], 'id') is not None or tdef[BaseType] == 'Array') else p_fname
            if tdef[BaseType] == 'Enumerated':      # Parse Enumerated Item
                pattern = fr'^{p_id}{p_fstr}{p_desc}$'
                if m := re.match(pattern, line):
                    return 'F', fielddef2jadn(int(m.group(1)), m.group(2), '', '', m.group(3) if m.group(3) else '')
            else:                                   # Parse Field
                pattern = f'^{p_id}{pn}{p_fstr}{p_range}{p_desc}$'
                if m := re.match(pattern, line):
                    m_range = '0..1' if m.group(5) else m.group(4)        # Convert 'optional' to range
                    fdesc = m.group(6) if m.group(6) else ''
                    return 'F', fielddef2jadn(int(m.group(1)), m.group(2), m.group(3), m_range if m_range else '', fdesc)
        else:
            raise_error(f'MARKDOWN Load - field with no type: {repr(line)}')

    return '', []


def markdown_loads(doc: str) -> dict:
    info = {}
    types = []
    fields = None
    for line in doc.splitlines():
        if line:
            t, v = line2jadn(line, types[-1] if types else None)    # Parse a MARKDOWN line
            if t == 'F':
                fields.append(v)
            elif fields:
                cleanup_tagid(fields)
                fields = None
            if t == 'M':
                info.update({v[0]: json.loads(v[1])})
            elif t == 'T':
                types.append(v)
                fields = types[-1][Fields]
    return {'info': info, 'types': types} if info else {'types': types}


def markdown_load(fname: Union[bytes, str, int]) -> dict:
    with open(fname, 'r') as f:
        return markdown_loads(f.read())


__all__ = [
    'markdown_dump',
    'markdown_dumps',
    'markdown_load',
    'markdown_loads',
    'markdown_style'
]



import copy
import json
import os

from collections import defaultdict
from typing import Dict, Optional, List, NoReturn, Set, TextIO, Tuple, Union
from ..core import check, load_any
from ..definitions import (
    TypeName, BaseType, TypeOptions, TypeDesc, Fields, FieldType, FieldOptions, OPTION_ID, is_builtin
)
from ..utils import build_deps, raise_error


class SchemaPackage:
    source: str                           # Filename or URL
    package: str                          # Namespace unique name
    schema: dict                          # JADN data
    namespaces: dict                      # Copy of meta['namespaces'] or empty {}
    tx: Dict[str, list]                   # Type index: {type name: type definition in schema}
    deps: Dict[str, Set[str]]            # Internal dependencies: {type1: {t2, t3}, type2: {t3, t4, t5}}
    refs: Dict[str, Dict[str, Set[str]]]  # External references {namespace1: {type1: {t2, t3}, ...}}
    used: Set[str]                        # Types from this package that have been referenced {t2, t3}

    def __init__(self, source: Union[dict, TextIO]):     # Read schema data, get package name
        if isinstance(source, dict):      # If schema is provided, save data
            self.schema = source
            self.source = ''
        else:
            self.schema = load_any(source)
            self.source = source.name

        try:
            self.package = self.schema['info']['package']
        except KeyError:
            raise_error(f'Schema package {self.source} must have a package ID')

        self.namespaces = self.schema['info']['namespaces'] if 'namespaces' in self.schema['info'] else {}
        self.clear()

    def load(self) -> NoReturn:     # Validate schema, build type dependencies and external references
        if hasattr(self, 'deps'):           # Ignore if already loaded
            return
        check(self.schema)
        self.tx = {t[TypeName]: t for t in self.schema['types']}
        self.deps = build_deps(self.schema)
        self.refs = defaultdict(lambda: defaultdict(set))
        for tn in self.deps:
            for dn in list(self.deps[tn]):     # Iterate over copy so original can be modified safely
                if ':' in dn:
                    self.deps[tn].remove(dn)
                    nsid, typename = dn.split(':', maxsplit=1)
                    try:
                        self.refs[self.namespaces[nsid]][tn].add(typename)
                    except KeyError as e:
                        raise_error(f'Resolve: no namespace defined for {e}')

    def clear(self) -> NoReturn:
        self.used = set()

    def add_used(self, t) -> NoReturn:
        self.used.add(t)


# Resolve util functions
def merge_tname(tref: str, package: str, namespaces: Dict[str, str], nsids: dict, sys: str = '$') -> str:
    """
    Convert reference to an imported type (nsid:TypeName) to a local type. Return unchanged if local.
    :param tref: Type name to be merged into the base package
    :param package: Namespace (unique name) of base package
    :param namespaces: Dict that maps base package's namespace ids to namespaces
    :param nsids: Dict that maps each namespace to a namespace id. If blank, do not append $nsid qualifier
    :param sys: single character system generated type delimiter
    :return: Local type name, qualified (TypeName$nsid) or unqualified (TypeName)
    """
    nsid, tname = tref.split(':', maxsplit=1) if ':' in tref else ('', tref)
    ns_id = nsids[namespaces[nsid] if nsid else package][0]
    return f'{tname}{sys}{ns_id}' if ns_id and not is_builtin(tname) else tname


def merge_typedef(tdef: list, package: str, namespaces: Dict[str, str], nsids: dict, sys: str = '$') -> list:
    oids = [OPTION_ID['ktype'], OPTION_ID['vtype'], OPTION_ID['enum'], OPTION_ID['pointer'], OPTION_ID['and']]  # Options whose value is/has a type name

    def update_opts(opts: List[str]) -> List[str]:
        return [f'{x[0]}{merge_tname(x[1:], package, namespaces, nsids, sys)}' if x[0] in oids else x for x in opts]

    td = [
        merge_tname(tdef[TypeName], package, namespaces, nsids, sys),
        tdef[BaseType],
        update_opts(tdef[TypeOptions]),
        tdef[TypeDesc]
    ]
    if len(tdef) > Fields:
        new_fields = copy.deepcopy(tdef[Fields])
        if td[BaseType] != 'Enumerated':
            for f in new_fields:
                f[FieldOptions] = update_opts(f[FieldOptions])
                f[FieldType] = merge_tname(f[FieldType], package, namespaces, nsids, sys)
        td.append(new_fields)
    return td


def make_enum(sm: SchemaPackage, tname: str, sys: str = '$') -> bool:
    if tname[0] in (OPTION_ID['enum'], OPTION_ID['pointer']):
        tn = tname[1:]
        if tn not in sm.used and tn in sm.tx:
            etype = 'Enum' if tname[0] == OPTION_ID['enum'] else 'Point'
            sm.schema['types'].append([f'{tn}{sys}{etype}', 'Enumerated', [], '', [[f[0], f[1], ''] for f in sm.tx[tn][Fields]]])
        return True
    return False


# add referenced typenames in this package to used list
def add_types(sm: SchemaPackage, tname: str, sys: str = '$') -> NoReturn:
    if {tname} - sm.used:
        sm.add_used(tname)
        try:
            for tn in sm.deps[tname]:
                add_types(sm, tn, sys)
        except KeyError as e:
            if not make_enum(sm, tname, sys):
                raise_error(f'Resolve: {e} not defined in {sm.package} ({sm.source})')


# add referenced types from other packages to used list
def resolve(sm: SchemaPackage, types: Set[str], packages: dict, sys: str = '$') -> NoReturn:
    if set(types) - sm.used:
        sm.load()
        for tn in types:
            add_types(sm, tn, sys)
        for pkg in sm.refs:
            if pkg in packages:
                print(f'  Resolve {pkg} into {sm.package}')
                resolve(packages[pkg], {t for k, v in sm.refs[pkg].items() if k in sm.used for t in v}, packages)
            else:
                print(f'* Resolve: package {pkg} not found.')


# Add referenced types to schema. dirname => other schema files
def resolve_imports(schema: dict, dirname: str, no_nsid: Tuple[str, ...] = ()):
    sys = '$'  # Character reserved for use in tool-generated type names
    # if 'namespaces' not in schema['info']:
    #    return schema
    root = SchemaPackage(schema)
    packages = {root.package: root}
    nsids = defaultdict(list)

    for fn in (os.path.join(dirname, f) for f in os.listdir(dirname) if os.path.splitext(f)[1] in ('.jadn', '.jidl')):
        with open(fn, 'r', encoding='utf-8') as fp:
            sm = SchemaPackage(fp)
        if sm.package not in packages:            # Add new package to list
            packages.update({sm.package: sm})
        elif root.package == sm.package and root.schema == sm.schema:     # Update source of root schema if found
            packages[sm.package].source = fn
        elif packages[sm.package].source != fn:                   # Flag multiple files with same package name
            print(f'* Duplicate package {sm.package}, Using: {packages[sm.package].source}, Ignoring: {fn}')
        for i, m in sm.namespaces.items():
            nsids[m].append('' if i in no_nsid else i)
    resolve(root, root.schema['info']['exports'] if 'exports' in root.schema['info'] else set(), packages)

    for t in root.used.copy():
        if t[0] in (OPTION_ID['enum'], OPTION_ID['pointer']):
            if t[1:] not in root.used:
                raise_error(f'Resolve: no base type for {t}')
            root.used.remove(t)     # Don't need explicit type if base type is present

    # Copy all needed types from other packages into root
    nsids[root.package] = ['']
    sc = {'info': {k: v for k, v in root.schema['info'].items() if k != 'namespaces'}, 'types': []}    # Remove namespaces
    for sm in [root] + [m for m in packages.values() if m.package != root.package]:
        sc['types'] += [merge_typedef(t, sm.package, sm.namespaces, nsids, sys) for t in sm.schema['types'] if t[TypeName] in sm.used]
    return sc



from .html import html_dump, html_dumps, html_load, html_loads
from .jidl import jidl_dump, jidl_dumps, jidl_load, jidl_loads, jidl_style
from .proto import proto_dump, proto_dumps, proto_load, proto_loads, proto_style
from .markdown import markdown_dump, markdown_dumps, markdown_style
from .dot_w import dot_dump, dot_dumps, dot_style
from .plant_w import plant_dump, plant_dumps, plant_style
from .jas_w import jas_dump, jas_dumps



"""
Translate a JADN schema into a GraphViz (https://graphviz.org/) graph display file
"""

import re
from datetime import datetime
from typing import NoReturn
from ..definitions import (TypeName, BaseType, TypeDesc, PRIMITIVE_TYPES,
                           Fields, FieldName, FieldType, FieldOptions, FieldDesc)
from ..utils import ftopts_s2d, multiplicity_str


# Wrap typenames at word boundaries to minimize node width, using a max of "lines" lines.
def wrapstr(ss: str, lines: int=3) -> str:
    p = 0
    bp = len(ss)/lines
    wrapped = ''
    for m in re.findall(r'([A-Z][a-z0-9]+)|([A-Z]+)|([a-z]+)', ss):  # TODO: update regex to support more word formats
        w = ''.join(m)
        if p > 0 and p + len(w)/2 > bp:
            wrapped += '\\n'
            bp += len(ss)/lines
        wrapped += w
        p += len(w)
    return wrapped


def dot_style() -> dict:
    # Return default generation options and GraphViz style attributes
    return {
        'detail': 'conceptual',     # Level of detail: conceptual, logical, information
        'links': True,              # Show link edges (dashed)
        'attributes': False,        # Show node attributes connected to entities (ellipse)
        'attr_color': 'palegreen',  # Attribute ellipse fill color
        'edge_label': True,         # Show field name on edges
        'multiplicity': True,       # Show multiplicity on edges
        'header': {                 # Options defined in GraphViz "Node, Edge and Graph Attributes"
            'graph': {
                'fontname': 'Times',
                'fontsize': 12
            },
            'node': {
                'fontname': 'Arial',
                'fontsize': 8,
                'shape': 'box',
                'style': 'filled',
                'fillcolor': 'lightskyblue1'
            },
            'edge': {
                'fontname': 'Arial',
                'fontsize': 7,
                'arrowsize': 0.5,
                'labelangle': 45.0,
                'labeldistance': 0.9
            },
            'bgcolor': 'transparent',
        }
    }


def dot_header(style: dict) -> str:
    header = 'digraph G {\n'
    for kw, val in style.items():
        v2 = f'="{val}"'
        if isinstance(val, dict):
            v2 = f' [{", ".join([k3 + "=" + str(v3) for k3, v3 in val.items()])}]'
        header += f'  {kw}{v2};\n'
    return header


def dot_dumps(schema: dict, style: dict=None) -> str:
    """
    Convert JADN schema to GraphViz "dot" file
    """
    s = dot_style()
    if style:
        s.update(style)

    text = ''
    for k, v in schema.get('info', {}).items():
        text += f'# {k}: {v}\n'
    text += f'\n{dot_header(s.get("header", ""))}\n'

    atypes = (*PRIMITIVE_TYPES, 'Enumerated')
    nodes = {tdef[TypeName]: k for k, tdef in enumerate(schema['types']) if tdef[BaseType] not in atypes}
    for td in schema['types']:
        node_type = f', shape="ellipse", fillcolor="{s["attr_color"]}"' if td[BaseType] in atypes else ''
        if s['attributes'] or not node_type:
            node_type = ', shape="hexagon"' if '<->' in td[TypeDesc] else node_type     # TODO: replace SOSA hacks
            text += f'  n{nodes[td[TypeName]]} [label="{wrapstr(td[TypeName])}"{node_type}]\n'
            for fd in td[Fields]:
                fopts, topts = ftopts_s2d(fd[FieldOptions])
                fieldtype = topts['vtype'] if fd[FieldType] == 'MapOf' else fd[FieldType]
                if fieldtype in nodes:
                    edge_attrs = ['style="dashed"'] if 'link' in fopts or '<=' in fd[FieldDesc] else []
                    if s['links'] or not edge_attrs:
                        edge_attrs += [f'label="{fd[FieldName]}"'] if s['edge_label'] else []
                        if s['multiplicity']:
                            opts_f = multiplicity_str(ftopts_s2d(fd[FieldOptions])[0])
                            opts_r = m.group(1) if (m := re.search(r'\[([^\]]+)\]', fd[FieldDesc])) else '1'
                            edge_attrs += [f'headlabel="{opts_f}", taillabel="{opts_r}"']
                        edge_list = ', '.join(edge_attrs)
                        edge = f' [{edge_list}]' if edge_list else ''
                        text += f'    n{nodes[td[TypeName]]} -> n{nodes[fieldtype]}{edge}\n'
    text += '}'
    return text


def dot_dump(schema: dict, fname: str, source: str = '', style: dict=None) -> NoReturn:
    with open(fname, 'w') as f:
        if source:
            f.write(f'"# Generated from {source}, {datetime.ctime(datetime.now())}"\n\n')
        f.write(dot_dumps(schema, style) + '\n')


__all__ = [
    'dot_dump',
    'dot_dumps',
    'dot_style'
]



"""
Translate JADN to JADN Interface Definition Language
"""
import json
import re

from datetime import datetime
from typing import NoReturn, TextIO, Tuple, Union
from ..definitions import TypeName, BaseType, TypeOptions, TypeDesc, Fields, ItemID, FieldID, INFO_ORDER
from ..utils import cleanup_tagid, get_optx, fielddef2jadn, jadn2fielddef, jadn2typestr, raise_error, typestr2jadn, etrunc

# JIDL -> JADN Type regexes
p_tname = r'\s*([-$\w]+)'               # Type Name
p_assign = r'\s*='                      # Type assignment operator
p_tstr = r'\s*(.*?)\s*\{?'              # Type definition
p_tdesc = r'(?:\s*\/\/\s*(.*?)\s*)?'    # Optional Type description

# JIDL -> JADN Field regexes
p_id = r'\s*(\d+)'  # Field ID
p_fname = r'\s+(\S+)' # Field Name
p_fstr = r'\s*(.*?)'  # Field definition or Enum value
p_range = r'\s*(?:\[([.*\w]+)\]|(optional))?'  # Multiplicity
p_desc = r'\s*(?:\/\/\s*(.*?)\s*)?'  # Field description, including field name if .id option


def jidl_style() -> dict:
    # Return default column positions
    return {
        'info': 12,     # Width of info name column (e.g., module:)
        'id': 4,        # Width of Field Id column
        'name': 16,     # Width of Field Name column
        'type': 35,     # Width of Field Type column
        'desc': None,   # Fixed-position descriptions - overrides type-dependent default if specified
        'page': None    # Truncate to specified page width if specified
    }


def jidl_dumps(schema: dict, style: dict = None) -> str:
    """
    Convert JADN schema to JADN-IDL

    :param dict schema: JADN schema
    :param dict style: Override default column widths if specified
    :return: JADN-IDL text
    :rtype: str
    """
    w = jidl_style()
    if style:
        w.update(style)   # Override any specified column widths

    text = ''
    info = schema['info'] if 'info' in schema else {}
    mlist = [k for k in INFO_ORDER if k in info]
    for k in mlist + list(set(info) - set(mlist)):              # Display info elements in fixed order
        text += f'{k:>{w["info"]}}: {json.dumps(info[k])}\n'    # TODO: wrap to page width, continuation-line parser

    wt = w['desc'] if w['desc'] else w['id'] + w['name'] + w['type']
    for td in schema['types']:
        tdef = f'{td[TypeName]} = {jadn2typestr(td[BaseType], td[TypeOptions])}'
        tdesc = '// ' + td[TypeDesc] if td[TypeDesc] else ''
        text += f'\n{tdef:<{wt}}{tdesc}'[:w['page']].rstrip() + '\n'
        idt = td[BaseType] == 'Array' or get_optx(td[TypeOptions], 'id') is not None
        for fd in td[Fields] if len(td) > Fields else []:       # TODO: constant-length types
            fname, fdef, fmult, fdesc = jadn2fielddef(fd, td)
            if td[BaseType] == 'Enumerated':
                fdesc = '// ' + fdesc if fdesc else ''
                fs = f'{fd[ItemID]:>{w["id"]}} {fname}'
                wf = w['id'] + w['name'] + 2
            else:
                fdef += '' if fmult == '1' else ' optional' if fmult == '0..1' else ' [' + fmult + ']'
                fdesc = '// ' + fdesc if fdesc else ''
                wn = 0 if idt else w['name']
                fs = f'{fd[FieldID]:>{w["id"]}} {fname:<{wn}} {fdef}'
                wf = w['id'] + w['type'] if idt else wt
            wf = w['desc'] if w['desc'] else wf
            text += etrunc(f'{fs:{wf}}{fdesc}'.rstrip(), w['page']) + '\n'
    return text


def jidl_dump(schema: dict, fname: Union[bytes, str, int], source='', style=None) -> NoReturn:
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write(f'/* Generated from {source}, {datetime.ctime(datetime.now())} */\n\n')
        f.write(jidl_dumps(schema, style))


# Convert JIDL to JADN
def line2jadn(line: str, tdef: list) -> Tuple[str, list]:
    if line.split('//', maxsplit=1)[0].strip():
        p_info = r'^\s*([-\w]+):\s*(.+?)\s*$'
        if m := re.match(p_info, line):
            return 'M', [m.group(1), m.group(2)]

        p_type = fr'^{p_tname}{p_assign}{p_tstr}{p_tdesc}$'
        if m := re.match(p_type, line):
            btype, topts, fo = typestr2jadn(m.group(2))
            assert fo == []                     # field options MUST not be included in typedefs
            newtype = [m.group(1), btype, topts, m.group(3) if m.group(3) else '', []]
            return 'T', newtype

        if tdef:        # looking for fields
            pn = '()' if (get_optx(tdef[TypeOptions], 'id') is not None or tdef[BaseType] == 'Array') else p_fname
            if tdef[BaseType] == 'Enumerated':      # Parse Enumerated Item
                pattern = fr'^{p_id}{p_fstr}{p_desc}$'
                if m := re.match(pattern, line):
                    return 'F', fielddef2jadn(int(m.group(1)), m.group(2), '', '', m.group(3) if m.group(3) else '')
            else:                                   # Parse Field
                pattern = f'^{p_id}{pn}{p_fstr}{p_range}{p_desc}$'
                if m := re.match(pattern, line):
                    m_range = '0..1' if m.group(5) else m.group(4)        # Convert 'optional' to range
                    fdesc = m.group(6) if m.group(6) else ''
                    return 'F', fielddef2jadn(int(m.group(1)), m.group(2), m.group(3), m_range if m_range else '', fdesc)
        else:
            raise_error(f'JIDL Load - field with no type: {repr(line)}')

    return '', []


def jidl_loads(doc: str) -> dict:
    info = {}
    types = []
    fields = None
    for line in doc.splitlines():
        if line:
            t, v = line2jadn(line, types[-1] if types else None)    # Parse a JIDL line
            if t == 'F':
                fields.append(v)
            elif fields:
                cleanup_tagid(fields)
                fields = None
            if t == 'M':
                info.update({v[0]: json.loads(v[1])})
            elif t == 'T':
                types.append(v)
                fields = types[-1][Fields]
    return {'info': info, 'types': types} if info else {'types': types}


def jidl_load(fp: TextIO) -> dict:
    return jidl_loads(fp.read())


__all__ = [
    'jidl_dump',
    'jidl_dumps',
    'jidl_load',
    'jidl_loads',
    'jidl_style'
]



"""
Translate JADN to Protobuf 3
"""
import json
import re

from datetime import datetime
from typing import NoReturn, Tuple, Union
from urllib.parse import urlparse
from ..definitions import TypeName, BaseType, TypeOptions, TypeDesc, Fields, INFO_ORDER,\
                          ItemID, FieldID, FieldName, FieldOptions, FieldDesc
from ..utils import cleanup_tagid, get_optx, fielddef2jadn, raise_error, typestr2jadn, topts_s2d, ftopts_s2d

# PROTO -> JADN Type regexes
p_tname = r'\s*([-$\w]+)'               # Type Name
p_assign = r'\s*='                      # Type assignment operator
p_tstr = r'\s*(.*?)\s*\{?'             # Type definition
p_tdesc = r'(?:\s*\/\/\s*(.*?)\s*)?'    # Optional Type description

# PROTO -> JADN Field regexes
p_id = r'\s*(\d+)'  # Field ID
p_fname = r'\s+([-:$\w]+\/?)?'  # Field Name with dir/ option (colon is deprecated, allow for now)
p_fstr = r'\s*(.*?)'  # Field definition or Enum value
p_range = r'\s*(?:\[([.*\w]+)\]|(optional))?'  # Multiplicity
p_desc = r'\s*(?:\/\/\s*(.*?)\s*)?'  # Field description, including field name if .id option


def proto_style() -> dict:
    # Return default column positions
    return {
        'comment_pre': True,     # If True, comments on separate line preceding type or field.  If false, inline.
    }


# Convert URI to java-style reversed internet domain name
def uri_to_revid(uri: str) -> str:
    u = urlparse(uri)
    return '.'.join(u.netloc.split(':')[0].split('.')[::-1] + u.path.replace('.', '-').split('/')[1:])


# Convert java-style reversed domain to URI (specify the number of domain components, default=2)
def revid_to_uri(revid: str, hostlen: int=2) -> str:
    o = revid.split('.')
    return f'http://{".".join(o[:hostlen][::-1])}/{"/".join(o[hostlen:])}'


def proto_dumps(schema: dict, style: dict = None) -> str:
    """
    Convert JADN schema to Protobuf 3

    :param dict schema: JADN schema
    :param dict style: Override default column widths if specified
    :return: Protobuf text
    :rtype: str
    """
    w = proto_style()
    if style:
        w.update(style)   # Override any specified column widths

    text = 'syntax = "proto3";\n'
    info = schema['info'] if 'info' in schema else {}
    mlist = [k for k in INFO_ORDER if k in info]
    for k in mlist + list(set(info) - set(mlist)):              # Display info elements in fixed order
        if k == 'package':
            text += f'package {uri_to_revid(info[k])};\n'
        else:
            text += f'// {k:>{w["info"]}}: {json.dumps(info[k])}\n'  # TODO: wrap to page width, parse continuation

    for td in schema['types']:
        topts = topts_s2d(td[TypeOptions])
        if td[TypeDesc]:
            text += f'// {td[TypeDesc]}\n'
        if td[BaseType] in ('Record', 'Map', 'Array'):
            text += f'message {td[TypeName]} {{  // ${td[BaseType]} {topts}\n'
        elif td[BaseType] == 'Enumerated':
            text += f'enum {td[TypeName]} {{  // ${topts}\n'
        else:
            text += f'// ${td[TypeName]}({td[BaseType]}) {topts}\n'

        for fd in td[Fields] if len(td) > Fields else []:       # TODO: constant-length types
            fopts, ftopts = ftopts_s2d(fd[FieldOptions])
            if fd[FieldDesc]:
                text += f'// {fd[FieldDesc]}\n'

        if td[BaseType] in ('Record', 'Map', 'Array', 'Enumerated', 'Choice'):
            text += '}\n\n'
    return text


def proto_dump(schema: dict, fname: Union[bytes, str, int], source='', style=None) -> NoReturn:
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write(f'/* Generated from {source}, {datetime.ctime(datetime.now())} */\n\n')
        f.write(proto_dumps(schema, style))


# Convert PROTO to JADN
def line2jadn(line: str, tdef: list) -> Tuple[str, list]:
    if line.split('//', maxsplit=1)[0].strip():
        p_info = r'^\s*([-\w]+):\s*(.+?)\s*$'
        if m := re.match(p_info, line):
            return 'M', [m.group(1), m.group(2)]

        p_type = fr'^{p_tname}{p_assign}{p_tstr}{p_tdesc}$'
        if m := re.match(p_type, line):
            btype, topts, fo = typestr2jadn(m.group(2))
            assert fo == []                     # field options MUST not be included in typedefs
            newtype = [m.group(1), btype, topts, m.group(3) if m.group(3) else '', []]
            return 'T', newtype

        if tdef:        # looking for fields
            pn = '()' if (get_optx(tdef[TypeOptions], 'id') is not None or tdef[BaseType] == 'Array') else p_fname
            if tdef[BaseType] == 'Enumerated':      # Parse Enumerated Item
                pattern = fr'^{p_id}{p_fstr}{p_desc}$'
                if m := re.match(pattern, line):
                    return 'F', fielddef2jadn(int(m.group(1)), m.group(2), '', '', m.group(3) if m.group(3) else '')
            else:                                   # Parse Field
                pattern = f'^{p_id}{pn}{p_fstr}{p_range}{p_desc}$'
                if m := re.match(pattern, line):
                    m_range = '0..1' if m.group(5) else m.group(4)        # Convert 'optional' to range
                    fdesc = m.group(6) if m.group(6) else ''
                    return 'F', fielddef2jadn(int(m.group(1)), m.group(2), m.group(3), m_range if m_range else '', fdesc)
        else:
            raise_error(f'PROTO Load - field with no type: {repr(line)}')

    return '', []


def proto_loads(doc: str) -> dict:
    info = {}
    types = []
    fields = None
    for line in doc.splitlines():
        if line:
            t, v = line2jadn(line, types[-1] if types else None)    # Parse a PROTO line
            if t == 'F':
                fields.append(v)
            elif fields:
                cleanup_tagid(fields)
                fields = None
            if t == 'M':
                info.update({v[0]: json.loads(v[1])})
            elif t == 'T':
                types.append(v)
                fields = types[-1][Fields]
    return {'info': info, 'types': types} if info else {'types': types}


def proto_load(fname: Union[bytes, str, int]) -> dict:
    with open(fname, 'r') as f:
        return proto_loads(f.read())


__all__ = [
    'proto_dump',
    'proto_dumps',
    'proto_load',
    'proto_loads',
    'proto_style'
]



"""
Translate a JADN schema into a PlantUML graph display file
"""

import re
from datetime import datetime
from typing import NoReturn
from ..definitions import (TypeName, BaseType, TypeDesc, PRIMITIVE_TYPES,
                           Fields, FieldID, FieldName, FieldType, FieldOptions, FieldDesc)
from ..utils import ftopts_s2d, multiplicity_str, jadn2fielddef


# Wrap typenames at word boundaries to minimize node width, using a max of "lines" lines.
def wrapstr(ss: str, lines: int = 3) -> str:
    p = 0
    bp = len(ss)/lines
    wrapped = ''
    for m in re.findall(r'([A-Z][a-z0-9]+)|([A-Z]+)|([a-z]+)', ss):  # TODO: update regex to support more word formats
        w = ''.join(m)
        if p > 0 and p + len(w)/2 > bp:
            wrapped += '\\n'
            bp += len(ss)/lines
        wrapped += w
        p += len(w)
    return wrapped


def plant_style() -> dict:
    # Return default generation options and PlantUML style attributes
    return {
        'detail': 'conceptual',     # Level of detail: conceptual, logical, information
        'links': True,              # Show link edges (dashed)
        'link_horizontal': True,    # Use e-w links vs. n-s containers
        'edge_label': True,         # Show field name on edges
        'multiplicity': True,       # Show multiplicity on edges
        'attributes': False,        # Show node attributes connected to entities (ellipse)
        'attr_color': 'palegreen',  # Attribute ellipse fill color
        'enums': 10,                # Show Enumerated values with max length (0: none, <0: unlimited)
        'header': [                 # PlantUML Options
            '\' !theme spacelab',
            'hide empty members',
            'hide circle'
        ]
    }


def plant_dumps(schema: dict, style: dict = None) -> str:
    """
    Convert JADN schema to PlantUML file
    """
    s = plant_style()
    if style:
        s.update(style)

    text = '@startuml\n'
    for k, v in schema.get('info', {}).items():
        text += f"' {k}: {v}\n"
    text += '\n' + '\n'.join(s.get('header', [])) + '\n\n'

    atypes = (*PRIMITIVE_TYPES, 'Enumerated')
    nodes = {tdef[TypeName]: k for k, tdef in enumerate(schema['types']) if tdef[BaseType] not in atypes}
    edges = ''
    for td in schema['types']:
        if (tn := td[TypeName]) in nodes:
            text += f'class "{tn}" as n{nodes[tn]} <<{td[BaseType]}>>\n'
            for fd in td[Fields]:
                fopts, ftopts = ftopts_s2d(fd[FieldOptions])
                fieldtype = ftopts['vtype'] if fd[FieldType] == 'MapOf' else fd[FieldType]
                if fieldtype in nodes:
                    rel = ('.' if 'link_horizontal' in s else '..') if 'link' in fopts else '--'
                    elabel = ' : ' + fd[FieldName] if s['edge_label'] else ''
                    mult_f, mult_r = '', ''
                    if s['multiplicity']:
                        mult_f = ' "' + multiplicity_str(fopts) + '"'
                        mult_r = '"1 "'
                    if s['detail'] == 'logical':
                        text += f'  n{nodes[tn]} : {fd[FieldName]}\n'
                    edges += f'  n{nodes[tn]} {mult_r}{rel}>{mult_f} n{nodes[fieldtype]}{elabel}\n'
                if s['detail'] == 'information':
                    fname, fdef, fmult, fdesc = jadn2fielddef(fd, td)
                    fdef += '' if fmult == '1' else ' [' + fmult + ']'
                    fdef = fdef.translate(str.maketrans({'(': '{', ')': '}'}))  # PlantUML parses parens as methods
                    text += f'  n{nodes[tn]} : {fd[FieldID]} {fname} : {fdef}\n'
    return text + edges + '@enduml'


def plant_dump(schema: dict, fname: str, source: str = '', style: dict = None) -> NoReturn:
    with open(fname, 'w') as f:
        if source:
            f.write(f'\' Generated from {source}, {datetime.ctime(datetime.now())}"\n\n')
        f.write(plant_dumps(schema, style) + '\n')


__all__ = [
    'plant_dump',
    'plant_dumps',
    'plant_style'
]



"""
Translate JADN to JSON Schema
"""
import json
import re

from datetime import datetime
from typing import Callable, Optional, Union
from ..definitions import (
    TypeName, BaseType, TypeOptions, TypeDesc, Fields, ItemID, ItemValue, ItemDesc, FieldID, FieldName, FieldType,
    FieldOptions, FieldDesc, OPTION_ID, is_builtin
)
from ..transform.transform import get_enum_items
from ..utils import dmerge, topts_s2d, ftopts_s2d, get_config


# Consts
JADN_FMT = {
    'x': {'contentEncoding': 'base16'},
    # 'eui': {'pattern': r'^([0-9a-fA-F]{2}[:-]){5}[0-9a-fA-F]{2}$'},
    'eui': {'pattern': r'^([0-9a-fA-F]{2}[:-]){5}[0-9A-Fa-f]{2}(([:-][0-9A-Fa-f]{2}){2})?$'},
    'ipv4-addr': {'pattern': r'^((25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])$'},
    'ipv6-addr': {'pattern': r'^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))(%.+)$'},
    'ipv4-net': {'pattern': r'^((25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9]?[0-9])(\/(3[0-2]|[0-2]?[0-9]))?$'},
    'ipv6-net': {'pattern': r'^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))(%.+)?s*(\/([0-9]|[1-9][0-9]|1[0-1][0-9]|12[0-8]))?$'},
    'i8': {'minimum': -128, 'maximum': 127},
    'i16': {'minimum': -32768, 'maximum': 32767},
    'i32': {'minimum': -2147483648, 'maximum': 2147483647}
}

CONFIG_MAX = {
    'Binary': '$MaxBinary',
    'String': '$MaxString',
    'Array': '$MaxElements',
    'ArrayOf': '$MaxElements',
    'Map': '$MaxElements',
    'MapOf': '$MaxElements',
    'Record': '$MaxElements'
}


# Util Functions
def get_items(stype: str, ctx: dict) -> Optional[list]:  # pylint: disable=R1710
    """
    return enumerated items if stype is Enumerated
    """
    td = ctx['type_defs']
    et = stype[1:] if stype[0] == OPTION_ID['enum'] else stype
    if et in td and td[et][BaseType] == 'Enumerated':
        to = topts_s2d(td[et][TypeOptions])
        et = to['enum'] if 'enum' in to else et

    to = topts_s2d(td[et][TypeOptions])
    if et in td and td[et][BaseType] in ('Enumerated', 'Array', 'Choice', 'Map', 'Record') and 'id' not in to:
        return [f[ItemValue] for f in td[et][Fields]]


def spaces(s: str) -> str:
    """
    Return name with dash and underscore replaced with space
    """
    return re.sub(r'[\-_]', ' ', s)


def fieldname(name: str) -> str:
    """
    Return TypeName converted to FieldName
    """
    return name.lower().replace('-', '_')


def pattern(vals: list) -> str:
    """
    Return a regex string from a list of values
    """

    return f"^({'|'.join(map(str, vals))})$"


# === Return JSON Schema keywords
def w_td(tname: str, desc: str) -> dict:
    """
    Make type and type description
    """
    return dmerge(
        {'type': tname},
        {'description': desc} if desc else {}
    )


def w_def(typ: str) -> dict:
    """
    Make definition reference
    """
    return {'#ref': f'#/definitions/{typ}'}


def w_kvtype(stype: str, ctx: dict) -> dict:
    """
    Make definition from ktype or vtype option
    """
    stypes = {'Boolean': 'boolean', 'Integer': 'integer', 'Number': 'number', 'String': 'string'}
    if stype in stypes:
        return {'type': stypes[stype]}
    if stype[0] in (OPTION_ID['enum'], OPTION_ID['pointer']):
        tdef = ctx['type_defs'][stype[1:]]
        topts = topts_s2d(tdef[TypeOptions])
        fields = get_enum_items(tdef, topts, ctx['type_defs'])
        idopt = 'id' in topts
        return w_enum(fields, FieldID if idopt else FieldName, FieldDesc, idopt, ctx)
    return {'$ref': f'#/definitions/{stype}'}


def w_enum(fields: list, vcol: int, dcol: int, idopt: bool, ctx: dict) -> dict:
    def fdesc(fld: list, desc: str, idopt: bool) -> str:
        """
        Make field description
        """
        return f'{fld[FieldName]} - {desc}' if idopt else desc

    values = [f[vcol] for f in fields]
    es = ctx['enum_style']
    assert es in ['enum', 'const', 'regex']
    if es == 'regex':
        return {'type': 'string', 'pattern': pattern(values)}
    if es == 'const':
        return {'oneOf': [{'const': f[vcol], 'description': fdesc(f, f[dcol], idopt)} for f in fields]}
    return {'enum': values}


def w_export(exp: list, ctx: dict) -> dict:
    """
    Make top-level definition header
    """
    if len(exp) == 1:
        return {'$ref': f'#/definitions/{exp[0]}'}
    return {
        'type': 'object',
        'additionalProperties': False,
        'properties': {fieldname(ctx['type_defs'][t][TypeName]): {'$ref': f'#/definitions/{t}'} for t in exp}
    }


def w_ref(tname: str, ctx: dict) -> dict:
    nsid, tn = tname.split(':', maxsplit=1) if ':' in tname else [None, tname]
    assert not is_builtin(tn)
    if nsid:
        imp = {'$ref': f"{ctx['info_imps'][nsid]}/definitions/{tn}"} if ctx['import_style'] == 'ref' and ctx['info_imps'] is not None else {}
        ctx['imported_types'][nsid].update({tn: imp})
        return {'$ref': f'#/imports/{nsid}/{tn}'}
    return {'$ref': f'#/definitions/{tn}'}


def w_fdef(f: list, ctx: dict) -> dict:
    """
    Make field definition
    """
    fopts, topts = ftopts_s2d(f[FieldOptions])
    if is_builtin(f[FieldType]):
        t = w_type(['', f[FieldType], [], f[FieldDesc]], topts, ctx)
    else:
        t = dmerge(w_ref(f[FieldType], ctx), {'description': f[FieldDesc]})

    minv = max(fopts.get('minc', 1), 1)
    maxv = fopts.get('maxc', 1)
    return t if minv <= 1 and maxv == 1 else dmerge(
        {'type': 'array'},
        {'description': f[FieldDesc]},
        {'uniqueItems': True} if 'unique' in topts else {},
        {'minItems': minv} if minv != 0 else {},
        {'maxItems': maxv} if maxv != 0 else {},
        {'items': t}
    )


def w_format(fmt: str) -> dict:
    """
    Make semantic validation keywords
    """
    if fmt:
        if jadn_fmt := JADN_FMT.get(fmt, None):
            return jadn_fmt
        return {'format': fmt}
    return {}


# ========== JADN Types ==========:
def t_binary(tdef: list, topts: dict, ctx: dict) -> dict:
    return dmerge(
        w_td('string', tdef[TypeDesc]),
        w_format(topts['format']) if 'format' in topts else {'contentEncoding': 'base64url'},
        # TODO: Fixme: JSON Schema cannot express length of content-encoded data
        # This would require adjusting string length by 2 for hex and 4/3 for base64
        # Impossible to calculate string length for other string formats
        # {'minLength': topts['minv']} if 'minv' in topts and topts['minv'] > 0 else {},
        # {'maxLength': topts['maxv']} if 'maxv' in topts else {}
    )


def t_boolean(tdef: list, topts: dict, ctx: dict) -> dict:
    return w_td('boolean', tdef[TypeDesc])


def t_integer(tdef: list, topts: dict, ctx: dict) -> dict:
    return dmerge(
        w_td('integer', tdef[TypeDesc]),
        {'minimum': topts['minv']} if 'minv' in topts else {},
        {'maximum': topts['maxv']} if 'maxv' in topts else {}
    )


def t_number(tdef: list, topts: dict, ctx: dict) -> dict:
    return dmerge(
        w_td('number', tdef[TypeDesc]),
        {'minimum': topts['minf']} if 'minf' in topts else {},
        {'maximum': topts['maxf']} if 'maxf' in topts else {}
    )


def t_null(tdef: list, topts: dict, ctx: dict) -> dict:
    return {}


def t_string(tdef: list, topts: dict, ctx: dict) -> dict:
    return dmerge(
        w_td('string', tdef[TypeDesc]),
        w_format(topts['format']) if 'format' in topts else {},
        {'minLength': topts['minv']} if 'minv' in topts and topts['minv'] != 0 else {},
        {'maxLength': topts['maxv']} if 'maxv' in topts else {},
        {'pattern': topts['pattern']} if 'pattern' in topts else {}
    )


def t_enumerated(tdef: list, topts: dict, ctx: dict) -> dict:
    item, tname = (ItemID, 'integer') if 'id' in topts else (ItemValue, 'string')
    fields = get_enum_items(tdef, topts, ctx['type_defs'])
    dcol = FieldDesc if 'enum' in topts or 'pointer' in topts else ItemDesc
    return dmerge(
        w_td(tname, tdef[TypeDesc]),
        w_enum(fields, item, dcol, 'id' in topts, ctx)
    )


def t_choice(tdef: list, topts: dict, ctx: dict) -> dict:
    return dmerge(
        w_td('object', tdef[TypeDesc]),
        {
            'additionalProperties': False,
            'minProperties': 1,
            'maxProperties': 1,
            'properties': {f[FieldName]: w_fdef(f, ctx) for f in tdef[Fields]}
        }
    )


def t_array(tdef: list, topts: dict, ctx: dict) -> dict:
    fmt = w_format(topts['format']) if 'format' in topts else {}
    if fmt:
        return dmerge(w_td('string', tdef[TypeDesc]), fmt)
    return dmerge(
        w_td('array', tdef[TypeDesc]),
        {'additionalItems': False},
        {'minItems': topts['minv']} if topts.get('minv', 0) != 0 else {},
        {'maxItems': topts['maxv']} if 'maxv' in topts else {},
        {'items': [w_fdef(f, ctx) for f in tdef[Fields]]}
    )


def t_array_of(tdef: list, topts: dict, ctx: dict) -> dict:
    return dmerge(
        w_td('array', tdef[TypeDesc]),
        {'uniqueItems': True} if 'unique' in topts else {},
        {'minItems': topts['minv']} if topts.get('minv', 0) != 0 else {},
        {'maxItems': topts['maxv']} if 'maxv' in topts else {},
        {'items': w_kvtype(topts['vtype'], ctx)}
    )


def t_map(tdef: list, topts: dict, ctx: dict) -> dict:
    def req(f: list) -> bool:
        fo = ftopts_s2d(f[FieldOptions])[0]
        return fo['minc'] >= 1 if 'minc' in fo else True

    required = [f[FieldName] for f in tdef[Fields] if req(f)]
    return dmerge(
        w_td('object', tdef[TypeDesc]),
        {'additionalProperties': False},
        {'required': required} if required else {},
        {'minProperties': topts['minv']} if topts.get('minv', 0) != 0 else {},
        {'maxProperties': topts['maxv']} if 'maxv' in topts else {},
        {'properties': {f[FieldName]: w_fdef(f, ctx) for f in tdef[Fields]}}
    )


def t_map_of(tdef: list, topts: dict, ctx: dict) -> dict:
    items = get_items(topts['ktype'], ctx)
    vtype = w_kvtype(topts['vtype'], ctx)
    return dmerge(
        w_td('object', tdef[TypeDesc]),
        {'additionalProperties': False},
        {'minProperties': topts['minv']} if topts.get('minv', 0) != 0 else {},
        {'maxProperties': topts['maxv']} if 'maxv' in topts else {},
        {'patternProperties': {pattern(items): vtype}} if items and ctx['enum_style'] == 'regex' else
        {'properties': {f: vtype for f in items}} if items else {}
    )


# Type Map Util
TYPE_WRITERS = {
    'Binary': t_binary,
    'Boolean': t_boolean,
    'Integer': t_integer,
    'Number': t_number,
    'String': t_string,
    'Enumerated': t_enumerated,
    'Choice': t_choice,
    'Array': t_array,
    'ArrayOf': t_array_of,
    'Map': t_map,
    'MapOf': t_map_of
}


def get_writer(btype: str, verbose=False) -> Callable[[list, dict, dict], dict]:
    if writer := TYPE_WRITERS.get(btype, None):
        return writer
    if btype == 'Record':
        return t_map if verbose else t_array
    raise TypeError(f'{btype} is not a valid base type')


def w_type(tdef: list, topts: dict, ctx: dict) -> dict:
    """
    Write a JADN type definition in JSON Schema format
    """
    if 'maxv' not in topts and tdef[BaseType] in CONFIG_MAX:
        topts['maxv'] = ctx['config'][CONFIG_MAX[tdef[BaseType]]]
    sc = get_writer(tdef[BaseType], ctx['verbose'])(tdef, topts, ctx)
    if 'and' in topts:
        return {'allOf': [sc, w_ref(topts['and'], ctx)]}
    if 'or' in topts:
        return {'anyOf': [sc, w_ref(topts['or'], ctx)]}
    return sc


# ========== Make JSON Schema ==========
# TODO: use string IDs if verbose=False
def json_schema_dumps(schema: dict, verbose=True, enum_style='enum', import_style='any'):
    #  verbose True: Record->object, field Names;  False: Record->array, field IDs
    assert enum_style in ('const', 'enum', 'regex')
    #  const: generate oneOf keyword with const for each item
    #  enum: generate an enum keyword containing all items
    #  regex: generate a regular expression that matches each item
    assert import_style in ('any', 'ref')
    #  any: ignore types defined in other modules, validate anything
    #  ref: generate $ref keywords that must be resolved before the JSON Schema can validate referenced types

    info = schema.get('info', {})
    td = {t[TypeName]: t for t in schema['types']}    # Build index of type definitions
    exports = [e for e in info.get('exports', []) if e in td]
    imported_types = {k: {} for k in info.get('namespaces', {})}
    ctx = {  # Translation context
        'config': get_config(info),
        'type_defs': td,
        'verbose': verbose,
        'enum_style': enum_style,
        'imported_types': imported_types,
        'import_style': import_style,
        'info_imps': info['imports'] if 'imports' in info else None
    }

    def tt(tdef: list, ctx: dict):  # Return type definition with title
        return dmerge(
            {'title': spaces(tdef[TypeName])},
            w_type(tdef, topts_s2d(tdef[TypeOptions]), ctx)
        )

    return json.dumps(dmerge(
        {'$schema': 'http://json-schema.org/draft-07/schema#'},
        {'$id': info['package']} if 'package' in info else {},
        {'title': info['title']} if 'title' in info else {},    # TODO: use items from INFO_ORDER
        {'version': info['version']} if 'version' in info else {},
        {'description': info['description']} if 'description' in info else {},
        {'comments': info['comments']} if 'comments' in info else {},
        {'copyright': info['copyright']} if 'copyright' in info else {},
        {'license': info['license']} if 'license' in info else {},
        w_export(exports, ctx),
        {'definitions': {t: tt(td[t], ctx) for t in (exports + [t for t in td if t not in exports])}},
        {'imports': imported_types} if imported_types else {}
    ), indent=2)


def json_schema_dump(jadn: dict, fname: Union[str, bytes, int], source=None, verbose=True, enum_style='enum', import_style='ref'):
    with open(fname, 'w') as f:
        if source:
            f.write(f'/* Generated from {source}, {datetime.ctime(datetime.now())} */\n\n')
        f.write(json_schema_dumps(jadn, verbose, enum_style, import_style))



"""
Abstract Object Encoder/Decoder

Object schema is specified in JSON Abstract Data Notation (JADN) format.

Codec currently supports three JSON concrete message formats (verbose,
concise, and minified) but can be extended to support XML or binary formats.

Copyright 2016, 2021 David Kemp
Licensed under the Apache License, Version 2.0
http://www.apache.org/licenses/LICENSE-2.0
"""
from typing import Any, Callable, Dict, List, Optional
from .codec import SymbolTableField, SymbolTableFieldDefinition, enctab, _decode_maprec, _encode_maprec
from .format_serialize_json import json_format_codecs, get_format_encode_function, get_format_decode_function
from .format_validate import format_validators, get_format_validate_function
from ..utils import ftopts_s2d, get_config, object_types, raise_error, topts_s2d
from ..definitions import (
    # Field Indexes
    BaseType, FieldID, FieldName,
    # Const values
    PRIMITIVE_TYPES, CORE_TYPES,
    # Dataclass
    TypeDefinition, GenFieldDefinition
)
from ..transform import unfold_extensions


class Codec:
    """
    Serialize (encode) and De-serialize (decode) values, validate against JADN syntax.
    verbose_rec - True: Record types encoded as maps
                 False: Record types encoded as arrays
    verbose_str - True: Identifiers encoded as strings
                 False: Identifiers encoded as integer tags
    """
    schema: dict  # better typing??
    config: dict  # better typing??
    format_validators: Dict[str, Dict[str, Callable[[Any], Any]]]
    format_codec: dict  # better typing??
    types: Dict[str, TypeDefinition]
    symtab = Dict[str, SymbolTableField]
    verbose_rec: bool
    verbose_str: bool

    def __init__(self, schema: dict, verbose_rec=False, verbose_str=False, config: dict = None):
        assert set(enctab) == set(CORE_TYPES)
        self.schema = unfold_extensions(schema)         # Convert extensions to core definitions
        conf = config if config else schema
        self.config = get_config(conf['info'] if 'info' in conf else None)
        self.format_validate = format_validators()      # Initialize format validation functions
        self.format_codec = json_format_codecs()        # Initialize format serialization functions
        # pre-index types to allow symtab forward refs
        self.types = {t.TypeName: t for t in object_types(self.schema['types'])}
        self.symtab = {}                         # Symbol table - pre-computed values for all datatypes
        self.set_mode(verbose_rec, verbose_str)  # Create symbol table based on encoding mode

    def decode(self, datatype: str, sval: Any) -> Any:  # Decode serialized value into API value
        try:
            ts = self.symtab[datatype]
        except KeyError:
            raise_error(f'Validation Error: Decode: datatype "{datatype}" is not defined')
        return ts.Decode(ts, sval, self)     # Dispatch to type-specific decoder

    def encode(self, datatype: str, aval: Any) -> Any:  # Encode API value into serialized value
        try:
            ts = self.symtab[datatype]
        except KeyError:
            raise_error(f'Validation Error: Encode: datatype "{datatype}" is not defined')
        return ts.Encode(ts, aval, self)     # Dispatch to type-specific encoder

    def set_mode(self, verbose_rec=False, verbose_str=False):
        # Build symbol table field entries
        def symf(fld: GenFieldDefinition, fa: int, fnames: dict) -> SymbolTableFieldDefinition:
            fo, to = ftopts_s2d(fld.FieldOptions)
            if to:
                raise_error(f'Validation Error: {fld.FieldName}: internal error: unexpected type options: {to}')
            fopts = {'minc': 1, 'maxc': 1, **fo}
            assert fopts['minc'] in (0, 1) and fopts['maxc'] == 1     # Other cardinalities have been simplified
            ctag: Optional[int] = None
            if 'tagid' in fopts:
                ctag = fopts['tagid'] if fa == FieldID else fnames[fopts['tagid']]
            return SymbolTableFieldDefinition(
                fld,        # SF_DEF: JADN field definition
                fopts,      # SF_OPT: Field options (dict)
                ctag        # SF_CTAG: tagid option
            )

        # Generate TypeRef pattern - concatenate NSID: and TypeName patterns
        def make_typeref_pattern(nsid: str, typename: str) -> dict:
            ns = nsid.lstrip('^').rstrip('$')
            tn = typename.lstrip('^').rstrip('$')
            return {'pattern': fr'^({ns}:)?{tn}$'}

        # Set configurable option values
        def config_opts(opts: List[str]) -> dict:
            op = [(v[0] + self.config[v[1:]]) if len(v) > 1 and v[1] == '$' else v for v in opts]
            return topts_s2d(op)

        def sym(t: TypeDefinition) -> SymbolTableField:  # Build symbol table based on encoding modes
            symval = SymbolTableField(
                t,                             # 0: S_TDEF:  JADN type definition
                enctab[t.BaseType].Enc,        # 1: S_ENCODE: Encoder for this type
                enctab[t.BaseType].Dec,        # 2: S_DECODE: Decoder for this type
                enctab[t.BaseType].eType,      # 3: S_ENCTYPE: Encoded value type
                config_opts(t.TypeOptions),    # 4: S_TOPTS:  Type Options (dict)
            )

            if t.BaseType == 'Record':
                symval.Encode = _encode_maprec   # if self.verbose_rec else _encode_array
                symval.Decode = _decode_maprec   # if self.verbose_rec else _decode_array
                symval.EncType = dict if self.verbose_rec else list
            if t.BaseType in ('Enumerated', 'Array', 'Choice', 'Map', 'Record'):
                fx = FieldName if 'id' not in symval.TypeOpts and t.BaseType != 'Array' and verbose_str else FieldID
                fa = FieldName if 'id' not in symval.TypeOpts else FieldID
                try:
                    symval.dMap = {f[fx]: f[fa] for f in t.Fields}
                    symval.eMap = {f[fa]: f[fx] for f in t.Fields}
                    fnames = {f[FieldID]: f[FieldName] for f in t.Fields}
                except IndexError as e:
                    raise IndexError(f'symval index error: {e}')
                if t.BaseType != 'Enumerated':
                    symval.Fld = {f[fx]: symf(f, fa, fnames) for f in t.Fields}
            if t.BaseType in ('Binary', 'String', 'Array', 'ArrayOf', 'Map', 'MapOf', 'Record'):
                minv = symval.TypeOpts.get('minv', 0)
                maxv = symval.TypeOpts.get('maxv', 0)
                if minv < 0 or maxv < 0:
                    raise_error(f'Validation Error: {t.TypeName}: length cannot be negative: {minv}..{maxv}')
                if maxv == 0:
                    maxv = self.config['$MaxElements']
                    if t.BaseType in ('Binary', 'String'):
                        maxv = self.config[f'$Max{t.BaseType}']
                symval.TypeOpts.update({'minv': minv, 'maxv': maxv})
            fmt = symval.TypeOpts.get('format', '')
            symval.FormatValidate = get_format_validate_function(self.format_validate, t.BaseType, fmt)
            symval.FormatEncode = get_format_encode_function(self.format_codec, t.BaseType, fmt)
            symval.FormatDecode = get_format_decode_function(self.format_codec, t.BaseType, fmt)
            return symval

        self.verbose_rec = verbose_rec
        self.verbose_str = verbose_str
        self.symtab = {t.TypeName: sym(t) for t in object_types(self.schema['types'])}
        if 'TypeRef' in self.types:
            self.symtab['TypeRef'].TypeOpts = make_typeref_pattern(self.config['$NSID'], self.config['$TypeName'])
        for t in PRIMITIVE_TYPES:
            self.symtab[t] = SymbolTableField(
                TypeDef=TypeDefinition('', t),
                Encode=enctab[t].Enc,
                Decode=enctab[t].Dec,
                EncType=enctab[t].eType,
                TypeOpts={},
                # TODO: check if t[BaseType] should just be t
                FormatValidate=get_format_validate_function(self.format_validate, t[BaseType], ''),
                FormatEncode=get_format_encode_function(self.format_codec, t[BaseType], ''),
                FormatDecode=get_format_decode_function(self.format_codec, t[BaseType], '')
            )


__all__ = ['Codec']



"""
 JADN Definitions

A JSON Abstract Data Notation (JADN) file contains a list of datatype definitions.  Each type definition
has a specified format - a list of four or five columns depending on whether the type is primitive or
structure: (name, base type, type options, type description [, fields]).

For the enumerated type each field definition is a list of three items: (tag, name, description).

For other structure types (array, choice, map, record) each field definition is a list of five items:
(tag, name, type, field options, field description).
"""
from copy import deepcopy
from dataclasses import Field, dataclass, field
from inspect import isfunction
from typing import List, Optional, Tuple, Union


class BasicDataclass:
    __annotations__: dict
    __default__: dict
    __keyindex__: Tuple[str, ...]

    def __init_subclass__(cls, **kwargs):
        cls.__keyindex__ = tuple(cls.__annotations__)
        cls.__default__ = {}
        for k in cls.__keyindex__:
            v = getattr(cls, k, None)
            if isinstance(v, Field):
                if isfunction(v.default):
                    cls.__default__[k] = v.default()
                elif isfunction(v.default_factory):
                    cls.__default__[k] = v.default_factory()
                else:
                    cls.__default__[k] = None
            else:
                cls.__default__[k] = v if isinstance(v, (int, float, str)) else deepcopy(v)

    def __getitem__(self, key: Union[int, slice, str]):
        if isinstance(key, slice):
            return [self[k] for k in self.__keyindex__[key]]
        if isinstance(key, int):
            key = list(self.__keyindex__)[key]
        return object.__getattribute__(self, key)

    def __setitem__(self, key: Union[int, str], val: any):
        if isinstance(key, int):
            key = list(self.__keyindex__)[key]
        return object.__setattr__(self, key, val)

    def __delitem__(self, key: Union[int, str]):
        if isinstance(key, int):
            key = list(self.__keyindex__)[key]
        object.__setattr__(self, self.__default__[key], self.__default__[key])

    def __len__(self):
        return len(self.__keyindex__)


# Datatype Definition columns
TypeName = 0            # Name of the type being defined
BaseType = 1            # Core type of the type being defined
TypeOptions = 2         # An array of zero or more TYPE_OPTIONS
TypeDesc = 3            # A non-normative description of the type
Fields = 4              # List of one or more items or fields

# Enumerated Item Definition columns
ItemID = 0              # Integer item identifier
ItemValue = 1           # String value of the item
ItemDesc = 2            # A non-normative description of the Enumerated item

# Field Definition columns
FieldID = 0             # Integer field identifier
FieldName = 1           # Name or label of the field
FieldType = 2           # Type of the field
FieldOptions = 3        # An array of zero or more FIELD_OPTIONS (and TYPE_OPTIONS if extended)
FieldDesc = 4           # A non-normative description of the field


# Dataclass Helpers
@dataclass
class EnumFieldDefinition(BasicDataclass):
    ItemID: int = 0
    ItemValue: str = ''
    ItemDesc: str = ''


@dataclass
class GenFieldDefinition(BasicDataclass):
    FieldID: int = 0
    FieldName: str = 'FieldName'
    FieldType: str = 'FieldType'
    FieldOptions: List[str] = field(default_factory=lambda: [])
    FieldDesc: str = ''


@dataclass
class TypeDefinition(BasicDataclass):
    TypeName: str = 'DefinitionName'
    BaseType: str = 'DefinitionType'
    TypeOptions: List[str] = field(default_factory=lambda: [])
    TypeDesc: str = ''
    Fields: Optional[Union[List[GenFieldDefinition], List[EnumFieldDefinition]]] = field(default_factory=lambda: [])


# Core datatypes
PRIMITIVE_TYPES = (
    'Binary',
    'Boolean',
    'Integer',
    'Number',
    'String',
)

SELECTOR_TYPES = (
    'Enumerated',       # enum option specifies fields derived from a defined type
    'Choice',
)

STRUCTURED_TYPES = (
    'Array',
    'ArrayOf',          # (value_type): instance is a container but definition has no fields
    'Map',
    'MapOf',            # (key_type, value_type): instance is a container but definition has no fields
    'Record',
)

CORE_TYPES = PRIMITIVE_TYPES + SELECTOR_TYPES + STRUCTURED_TYPES

FIELD_LENGTH = {
    'Binary': 0,
    'Boolean': 0,
    'Integer': 0,
    'Number': 0,
    'String': 0,
    'Enumerated': 3,    # 0 if Enumerated type definition contains enum or pointer option
    'Choice': 5,
    'Array': 5,
    'ArrayOf': 0,
    'Map': 5,
    'MapOf': 0,
    'Record': 5,
}


def is_builtin(t: str) -> bool:      # Is a core type
    return t in CORE_TYPES


def has_fields(t: str) -> bool:      # Is a type with fields listed in definition
    return FIELD_LENGTH[t] == 5 if is_builtin(t) else False


# Option Tags/Keys
#   JADN TypeOptions and FieldOptions contain a list of strings, each of which is an option.
#   The first character of an option string is the type ID; the remaining characters are the value.
#   The option string is converted into a Name: Value pair before use.
#   The tables list the unicode codepoint of the ID and the corresponding Name and value type.

TYPE_OPTIONS = {        # Option ID: (name, value type, canonical order) # ASCII ID
    61: ('id', lambda x: True, 1),          # '=', Enumerated type and Choice/Map/Record keys are ID not Name
    42: ('vtype', lambda x: x, 2),          # '*', Value type for ArrayOf and MapOf
    43: ('ktype', lambda x: x, 3),          # '+', Key type for MapOf
    35: ('enum', lambda x: x, 4),           # '#', enumeration derived from Array/Choice/Map/Record type
    62: ('pointer', lambda x: x, 5),        # '>', enumeration of pointers derived from Array/Choice/Map/Record type
    47: ('format', lambda x: x, 6),         # '/', semantic validation keyword, may affect serialization
    37: ('pattern', lambda x: x, 7),        # '%', regular expression that a string must match
    121: ('minf', float, 8),                # 'y', minimum Number value
    122: ('maxf', float, 9),                # 'z', maximum Number value
    123: ('minv', int, 10),                 # '{', minimum byte or text string length, Integer value, element count
    125: ('maxv', int, 11),                 # '}', maximum byte or text string length, Integer value, element count
    113: ('unique', lambda x: True, 12),    # 'q', ArrayOf instance must not contain duplicates
    115: ('set', lambda x: True, 13),       # 's', ArrayOf instance is unordered and unique
    98: ('unordered', lambda x: True, 14),  # 'b', ArrayOf instance is unordered and not unique (bag)
    88: ('extend', lambda x: True, 15),     # 'X', Type has an extension point where fields may be appended
    33: ('default', lambda x: x, 16),       # '!', Default or constant value of instances of this type
    0x2229: ('and', lambda x: x, 17),       # '∩', INTERSECTION - instance must also match referenced type (allOf)
    0x222a: ('or', lambda x: x, 18),        # '∪', UNION - instance must match at least one of the types (anyOf)
}

FIELD_OPTIONS = {
    91: ('minc', int, 19),                  # '[', minimum cardinality, default = 1, 0 = field is optional
    93: ('maxc', int, 20),                  # ']', maximum cardinality, default = 1, 0 = inherited max, not 1 = array
    38: ('tagid', int, 21),                 # '&', field that specifies the type of this field
    60: ('dir', lambda x: True, 22),        # '<', pointer enumeration treats field as a collection
    75: ('key', lambda x: True, 23),        # 'K', field is a primary key for this type
    76: ('link', lambda x: True, 24),       # 'L', field is a link (foreign key) to an instance of FieldType
}

OPTION_ID = {   # Pre-computed reverse index - MUST match TYPE_OPTIONS and FIELD_OPTIONS
    'id':       chr(61),
    'vtype':    chr(42),
    'ktype':    chr(43),
    'enum':     chr(35),
    'pointer':  chr(62),
    'format':   chr(47),
    'pattern':  chr(37),
    'minf':     chr(121),
    'maxf':     chr(122),
    'minv':     chr(123),
    'maxv':     chr(125),
    'unique':   chr(113),
    'set':      chr(115),
    'unordered': chr(98),
    'extend':   chr(88),
    'default':  chr(33),
    'and':      chr(0x2229),
    'or':       chr(0x222a),
    'minc':     chr(91),
    'maxc':     chr(93),
    'tagid':    chr(38),
    'dir':      chr(60),
    'key':      chr(75),
    'link':     chr(76),
}

REQUIRED_TYPE_OPTIONS = {
    'Binary': [],
    'Boolean': [],
    'Integer': [],
    'Number': [],
    'String': [],
    'Enumerated': [],
    'Choice': [],
    'Array': [],
    'ArrayOf': ['vtype'],
    'Map': [],
    'MapOf': ['ktype', 'vtype'],
    'Record': [],
}

ALLOWED_TYPE_OPTIONS = {
    'Binary': ['format', 'minv', 'maxv', 'and', 'or'],
    'Boolean': ['and', 'or'],
    'Integer': ['format', 'minv', 'maxv', 'and', 'or'],
    'Number': ['format', 'minf', 'maxf', 'and', 'or'],
    'String': ['format', 'pattern', 'minv', 'maxv', 'and', 'or'],
    'Enumerated': ['id', 'enum', 'pointer', 'extend', 'and', 'or'],
    'Choice': ['id', 'extend', 'and', 'or'],
    'Array': ['extend', 'format', 'minv', 'maxv', 'and', 'or'],
    'ArrayOf': ['vtype', 'minv', 'maxv', 'unique', 'set', 'unordered', 'and', 'or'],
    'Map': ['id', 'extend', 'minv', 'maxv', 'and', 'or'],
    'MapOf': ['ktype', 'vtype', 'minv', 'maxv', 'and', 'or'],
    'Record': ['extend', 'minv', 'maxv', 'and', 'or'],
}

# Ensure jsonschema prerequisite packages are installed, e.g., rfc3987 for uri/iri validation
FORMAT_JS_VALIDATE = {      # Semantic validation formats defined by JSON Schema 2019-09 Sec 7.3
    'date-time': 'String',
    'date': 'String',
    'time': 'String',
    'duration': 'String',
    # 'email': 'String',        # jsonschema package has deliberately buggy email - won't be fixed
    'idn-email': 'String',
    # 'hostname': 'String',     # jsonschema package needs bug fix
    'idn-hostname': 'String',
    'ipv4': 'String',           # doesn't allow netmask prefix length
    'ipv6': 'String',           # doesn't allow netmask prefix length
    'uri': 'String',
    'uri-reference': 'String',
    'iri': 'String',
    'iri-reference': 'String',
    'uuid': 'String',
    'uri-template': 'String',
    'json-pointer': 'String',
    'relative-json-pointer': 'String',
    'regex': 'String'
}

FORMAT_VALIDATE = {         # Semantic validation formats defined by JADN
    'email': 'String',          # Use this instead of jsonschema
    'hostname': 'String',       # Use this instead of jsonschema
    'eui': 'Binary',            # IEEE Extended Unique Identifier, 48 bits or 64 bits
    'ipv4-addr': 'Binary',      # IPv4 address as specified in RFC 791 Section 3.1
    'ipv6-addr': 'Binary',      # IPv6 address as specified in RFC 8200 Section 3
    'ipv4-net': 'Array',        # Binary IPv4 address and Integer prefix length, RFC 4632 Section 3.1
    'ipv6-net': 'Array',        # Binary IPv6 address and Integer prefix length, RFC 4291 Section 2.3
    'i8': 'Integer',            # Signed 8 bit integer [-128 .. 127]
    'i16': 'Integer',           # Signed 16 bit integer [-32768 .. 32767]
    'i32': 'Integer',           # Signed 32 bit integer [-2147483648 .. 2147483647]
    'i64': 'Integer',           # Signed 64 bit integer [-2^63 .. 2^63 -1]
    # 'u#': 'Integer',            # Unsigned '#'-bit integer or bit field where #>0, [0 .. 2^# -1]
}

FORMAT_SERIALIZE = {        # Data representation formats for one or more serializations
    'eui': 'Binary',            # IEEE EUI, 'hex-byte-colon' text representation, (e.g., 00:1B:44:11:3A:B7)
    'ipv4-addr': 'Binary',      # IPv4 'dotted-quad' text representation, RFC 2673 Section 3.2
    'ipv6-addr': 'Binary',      # IPv6 text representation, RFC 4291 Section 2.2
    'ipv4-net': 'Array',        # IPv4 Network Address CIDR text string, RFC 4632 Section 3.1
    'ipv6-net': 'Array',        # IPv6 Network Address CIDR text string, RFC 4291 Section 2.3
    'b': 'Binary',              # Base64url - RFC 4648 Section 5 (default text representation of Binary type)
    'x': 'Binary',              # Hex - base16 - lowercase out, case-folding in
    'X': 'Binary',              # Hex - RFC 4648 Section 8 - uppercase only
    'datetime-ms': 'Integer',   # Milliseconds from the epoch, RFC 3339 date-time text representation
    'i8': 'Integer',            # 8 bit field - these affect packed (RFC 791 style) serializations
    'i16': 'Integer',           # 16 bit field
    'i32': 'Integer',           # 32 bit field
    'i64': 'Integer',           # 64 bit field
    # 'u#': 'Integer',            # #-bit field
    'f16': 'Number',            # IEEE 754 Half-Precision Float - these affect CBOR serialization
    'f32': 'Number',            # IEEE 754 Single-Precision Float
    'f64': 'Number',            # IEEE 754 Double-Precision Float (default binary representation of Number type)
}

VALID_FORMATS = {**FORMAT_JS_VALIDATE, **FORMAT_VALIDATE, **FORMAT_SERIALIZE}

DEFAULT_CONFIG = {          # Configuration values to use if not specified in schema
    '$MaxBinary': 255,          # Maximum number of octets for Binary types
    '$MaxString': 255,          # Maximum number of characters for String types
    '$MaxElements': 100,        # Maximum number of items/properties for container types
    '$Sys': '$',                # System reserved character for TypeName
    '$TypeName': '^[A-Z][-$A-Za-z0-9]{0,31}$',   # Type Name regex
    '$FieldName': '^[a-z][_A-Za-z0-9]{0,31}$',   # Field Name regex
    '$NSID': '^[A-Za-z][A-Za-z0-9]{0,7}$',       # Namespace ID regex
    '$TypeRef': '^$'            # Placeholder.  Actual pattern is ($NSID ':')? $TypeName
}

EXTENSIONS = {
    'AnonymousType',            # TYPE_OPTIONS included in FieldOptions
    'Multiplicity',             # maxc other than 1, or minv other than 0 (optional) or 1 (required)
    'DerivedEnum',              # enum and pointer/dir options, create Enumerated type of fields or JSON Pointers
    'MapOfEnum',                # ktype option specifies an Enumerated type
    'Link',                     # key and link options
}

INFO_ORDER = ('title', 'package', 'version', 'description', 'comments',
              'copyright', 'license', 'namespaces', 'exports', 'config')    # Display order

GRAPH_DETAIL = ('conceptual', 'logical', 'information')

# Type Hinting
OPTION_TYPES = Union[int, float, str]



import base64
import binascii
import re
import string

from datetime import datetime, timezone
from ipaddress import IPv4Address, IPv6Address
from typing import Any, Callable, Dict, Tuple
from ..definitions import FORMAT_SERIALIZE

FormatFunction = Callable[[Any], Any]
FormatTable = Dict[str, Dict[str, Tuple[FormatFunction, FormatFunction]]]
# TODO: Convert a2s_ipvX_net and s2a_ipvX_net functions to use ipaddress.IPvXNetwork


def _format_pass(val: Any) -> Any:
    return val


def _codec_function(format_table: FormatTable, base_type: str, format_kw: str, direction: int) -> FormatFunction:
    try:
        if not format_kw:
            format_kw = {'Binary': 'b', 'Number': 'f64'}[base_type]  # Set default format for type if one exists
        return format_table[base_type][format_kw][direction]
    except KeyError:
        return _format_pass


# Return serialization functions for the specified format keyword and data type
def get_format_encode_function(format_table: FormatTable, base_type: str, format_kw: str) -> FormatFunction:
    return _codec_function(format_table, base_type, format_kw, 0)


def get_format_decode_function(format_table: FormatTable, base_type: str, format_kw: str) -> FormatFunction:
    return _codec_function(format_table, base_type, format_kw, 1)


# Binary to String, String to Binary conversion functions
def b2s_hex(bval: bytes) -> str:      # Convert from binary to hex string
    return base64.b16encode(bval).decode()


def s2b_hex(sval: str) -> bytes:      # Convert from hex string to binary
    try:
        return base64.b16decode(sval)
    except binascii.Error:
        raise TypeError


def b2s_hex_lc(bval: bytes) -> str:      # Convert from binary to hex string
    return base64.b16encode(bval).decode().lower()


def s2b_hex_lc(sval: str) -> bytes:      # Convert from hex string to binary
    try:
        return base64.b16decode(sval, casefold=True)
    except binascii.Error:
        raise TypeError


def b2s_base64url(bval: bytes) -> str:      # Convert from binary to base64url string
    return base64.urlsafe_b64encode(bval).decode().rstrip('=')


def s2b_base64url(sval: str) -> bytes:      # Convert from base64url string to binary
    v = sval
    if mod := len(sval) % 4:  # Pad b64 string out to a multiple of 4 characters
        v = f"{sval}{'=' * (4-mod)}"
    if set(v) - set(string.ascii_letters + string.digits + '-_='):  # Python 2 doesn't support Validate
        raise TypeError('base64decode: bad character')
    return base64.b64decode(str(v), altchars=b'-_')


def b2s_ipv4_addr(bval: bytes) -> str:      # Convert IPv4 address from binary to string
    return IPv4Address(bval).compressed


def b2s_ipv6_addr(bval: bytes) -> str:        # Convert ipv6 address from binary to string
    return IPv6Address(bval).compressed


def s2b_ipv4_addr(sval: str) -> bytes:    # Convert IPv4 addr from string to binary
    return IPv4Address(sval).packed


def s2b_ipv6_addr(sval: str) -> bytes:    # Convert IPv6 address from string to binary
    return IPv6Address(sval).packed


FORMAT_CONVERT_BINARY_FUNCTIONS = {
    'b': (b2s_base64url, s2b_base64url),            # Base64url
    'x': (b2s_hex_lc, s2b_hex_lc),                  # Hex lower case
    'X': (b2s_hex, s2b_hex),                        # Hex upper case (RFC conforming)
    'ipv4-addr': (b2s_ipv4_addr, s2b_ipv4_addr),    # IPv4 Address
    'ipv6-addr': (b2s_ipv6_addr, s2b_ipv6_addr),    # IPv6 Address
    'eui': (b2s_hex, s2b_hex),                      # EUI - TODO: write colon-hex A0:32:F9:...
}


# IP Net (address, prefix length tuple) conversions
def a2s_ipv4_net(aval: [str, int]) -> str:
    if aval[1] < 0 or aval[1] > 32:  # Verify prefix length is valid
        raise ValueError
    sa = b2s_ipv4_addr(aval[0])      # Convert Binary bytes to type-specific string
    return f'{sa}/{aval[1]}' if aval[1] != 32 else sa


def a2s_ipv6_net(aval: [str, int]) -> str:
    if aval[1] < 0 or aval[1] > 128:  # Verify prefix length is valid
        raise ValueError
    sa = b2s_ipv6_addr(aval[0])       # Convert Binary bytes to type-specific string
    return f'{sa}/{aval[1]}' if aval[1] != 128 else sa


def s2a_ipv4_net(sval: str) -> [str, int]:
    sv = sval.split('/', 1)
    sa = s2b_ipv4_addr(sv[0])                  # Convert type-specific string to Binary bytes
    prefix_len = int(sv[1]) if len(sv) > 1 else 32
    if prefix_len < 0 or prefix_len > 32:
        raise ValueError
    return [sa, prefix_len]


def s2a_ipv6_net(sval: str) -> [str, int]:
    sv = sval.split('/', 1)
    sa = s2b_ipv6_addr(sv[0])                  # Convert type-specific string to Binary bytes
    prefix_len = int(sv[1]) if len(sv) > 1 else 128
    if prefix_len < 0 or prefix_len > 128:
        raise ValueError
    return [sa, prefix_len]


FORMAT_CONVERT_MULTIPART_FUNCTIONS = {
    'ipv4-net': (a2s_ipv4_net, s2a_ipv4_net),       # IPv4 Net Address with CIDR prefix length
    'ipv6-net': (a2s_ipv6_net, s2a_ipv6_net),       # IPv6 Net Address with CIDR prefix length
}


def int2datems(dt: int) -> str:
    y = datetime.isoformat(datetime.fromtimestamp(dt/1000., timezone.utc))
    if m := re.match(r'^(.+)(\.\d\d\d)(\d\d\d)(.+)$', y):   # strip microseconds to milliseconds
        y = m.group(1) + m.group(2) + m.group(4)
    return y


def datems2int(dts: str) -> int:
    x = datetime.fromisoformat(dts.upper().replace('Z', '+00:00').replace(' ', 'T'))
    return int(1000 * datetime.timestamp(x))


# No special sized integer serialization in JSON.  Define these for packed encoding.
FORMAT_CONVERT_INTEGER_FUNCTIONS = {
    'datetime-ms': (int2datems, datems2int),      # RFC 3339 milliseconds from epoch
    'i8': (_format_pass, _format_pass),
    'i16': (_format_pass, _format_pass),
    'i32': (_format_pass, _format_pass),
    'i64': (_format_pass, _format_pass)
}

# No special serialization in JSON. Define these for CBOR encoding.
FORMAT_CONVERT_NUMBER_FUNCTIONS = {
    'f16': (_format_pass, _format_pass),
    'f32': (_format_pass, _format_pass),
    'f64': (_format_pass, _format_pass)
}


# Create a table listing the serialization functions for each format keyword
def json_format_codecs() -> Dict[str, Dict[str, Tuple[FormatFunction, FormatFunction]]]:  # Return table of JSON format serialization functions
    # Ensure code is in sync with JADN definitions
    assert set(FORMAT_SERIALIZE) == \
           set(FORMAT_CONVERT_BINARY_FUNCTIONS) |\
           set(FORMAT_CONVERT_MULTIPART_FUNCTIONS) |\
           set(FORMAT_CONVERT_INTEGER_FUNCTIONS) |\
           set(FORMAT_CONVERT_NUMBER_FUNCTIONS)

    return {
        'Binary': FORMAT_CONVERT_BINARY_FUNCTIONS,
        'Array': FORMAT_CONVERT_MULTIPART_FUNCTIONS,
        'Integer': FORMAT_CONVERT_INTEGER_FUNCTIONS,
        'Number': FORMAT_CONVERT_NUMBER_FUNCTIONS
    }



"""
**********************************
JSON Abstract Data Notation (JADN)
**********************************
"""
from .core import analyze, check, dump, dumps, load, loads, load_any, data_dir
from .utils import (
    topts_s2d, ftopts_s2d, opts_d2s, get_optx, del_opt, cleanup_tagid, opts_sort, canonicalize, multiplicity_str,
    raise_error, build_deps, get_config, jadn2typestr, typestr2jadn, jadn2fielddef, fielddef2jadn
)

import jadn.codec
import jadn.convert
import jadn.transform
import jadn.translate

__version__ = '0.6.17'
